"""Tests for provided template tags."""
from django.test import TestCase

from django_afip.templatetags.django_afip import format_cuit


class FormatCuitTagTestCase(TestCase):
    """Test the format_cuit tag."""

    def test_good_string_input(self):
        """Test valid string inputs."""
        self.assertEqual(format_cuit("20329642330"), "20-32964233-0")
        self.assertEqual(format_cuit("20-32964233-0"), "20-32964233-0")

    def test_good_numeric_input(self):
        """Test valid numerical input."""
        self.assertEqual(format_cuit(20329642330), "20-32964233-0")

    def test_bad_string_input(self):
        """Test invalid string input."""
        self.assertEqual(format_cuit("blah blah"), "blah blah")

    def test_bad_numeric_input(self):
        """Test invalid numerical input."""
        self.assertEqual(format_cuit(1234), 1234)



import pytest

from django_afip.factories import get_test_file


@pytest.fixture
def expired_crt() -> bytes:
    with open(get_test_file("test_expired.crt"), "rb") as crt:
        return crt.read()


@pytest.fixture
def expired_key() -> bytes:
    with open(get_test_file("test_expired.key"), "rb") as key:
        return key.read()



"""Tests for AFIP-WS related classes."""
from datetime import datetime
from datetime import timedelta
from unittest import skip
from unittest.mock import patch

import pytest
from django.core import management
from django.test import TestCase
from django.utils.timezone import now
from factory.django import FileField

from django_afip import exceptions
from django_afip import factories
from django_afip import models
from testapp.testmain.tests.testcases import LiveAfipTestCase
from testapp.testmain.tests.testcases import PopulatedLiveAfipTestCase


@pytest.mark.live
class AuthTicketTest(TestCase):
    """Test AuthTicket methods."""

    def test_bad_cuit(self):
        """Test using the wrong cuit for a key pair."""

        taxpayer = factories.AlternateTaxpayerFactory(cuit=20329642339)
        taxpayer.create_ticket("wsfe")

        with self.assertRaisesRegex(
            exceptions.AfipException,
            # Note: AFIP apparently edited this message and added a typo:
            "ValidacionDeToken: No apareci[oó] CUIT en lista de relaciones:",
        ):
            taxpayer.fetch_points_of_sales()

    def test_bogus_certificate_exception(self):
        """Test that using a junk ceritificates raises as expected."""

        # New TaxPayers will fail to save with an invalid cert, but many
        # systems may have very old TaxPayers, externally created, or other
        # stuff, so this scenario might still be possible.
        with patch(
            "django_afip.models.TaxPayer.get_certificate_expiration",
            spec=True,
            return_value=None,
        ):
            taxpayer = factories.TaxPayerFactory(
                key=FileField(data=b"Blah"),
                certificate=FileField(data=b"Blah"),
            )

        with self.assertRaises(exceptions.CorruptCertificate) as e:
            taxpayer.create_ticket("wsfe")

        self.assertNotIsInstance(e, exceptions.AfipException)

    def test_no_active_taxpayer(self):
        """Test that no TaxPayers raises an understandable error."""
        with self.assertRaisesMessage(
            exceptions.AuthenticationError,
            "There are no taxpayers to generate a ticket.",
        ):
            models.AuthTicket.objects.get_any_active("wsfe")

    def test_expired_certificate_exception(self):
        """Test that using an expired ceritificate raises as expected."""
        taxpayer = factories.TaxPayerFactory(
            key=FileField(from_path=factories.get_test_file("test_expired.key")),
            certificate=FileField(
                from_path=factories.get_test_file("test_expired.crt")
            ),
        )

        with self.assertRaises(exceptions.CertificateExpired):
            taxpayer.create_ticket("wsfe")

    def test_untrusted_certificate_exception(self):
        """
        Test that using an untrusted ceritificate raises as expected.
        """
        # Note that we hit production with a sandbox cert here:
        taxpayer = factories.TaxPayerFactory(is_sandboxed=False)

        with self.assertRaises(exceptions.UntrustedCertificate):
            taxpayer.create_ticket("wsfe")


class PopulationTest(LiveAfipTestCase):
    """
    Tests models population view.

    As a side effect, also test valid ticket generation.
    """

    def test_population_command(self):
        """Test the afipmetadata command."""
        management.call_command("afipmetadata")

        receipts = models.ReceiptType.objects.count()
        concepts = models.ConceptType.objects.count()
        documents = models.DocumentType.objects.count()
        vat = models.VatType.objects.count()
        tax = models.TaxType.objects.count()
        currencies = models.CurrencyType.objects.count()

        self.assertGreater(receipts, 0)
        self.assertGreater(concepts, 0)
        self.assertGreater(documents, 0)
        self.assertGreater(vat, 0)
        self.assertGreater(tax, 0)
        self.assertGreater(currencies, 0)

    def test_metadata_deserialization(self):
        """Test that we deserialize descriptions properly."""
        management.call_command("afipmetadata")

        # This asserting is tied to current data, but it validates that we
        # don't mess up encoding/decoding the value we get.
        # It _WILL_ need updating if the upstream value ever changes.
        fac_c = models.ReceiptType.objects.get(code=11)
        self.assertEqual(fac_c.description, "Factura C")


class TaxPayerTest(LiveAfipTestCase):
    """Test TaxPayer methods."""

    def test_fetch_points_of_sale(self):
        """Test the ``fetch_points_of_sales`` method."""
        taxpayer = models.TaxPayer.objects.first()
        taxpayer.fetch_points_of_sales()

        points_of_sales = models.PointOfSales.objects.count()
        self.assertGreater(points_of_sales, 0)


class ReceiptQuerySetTestCase(PopulatedLiveAfipTestCase):
    """Test ReceiptQuerySet methods."""

    def _good_receipt(self):
        receipt = factories.ReceiptFactory(
            point_of_sales=models.PointOfSales.objects.first(),
        )
        factories.VatFactory(vat_type__code=5, receipt=receipt)
        factories.TaxFactory(tax_type__code=3, receipt=receipt)
        return receipt

    def _bad_receipt(self):
        receipt = factories.ReceiptFactory(
            point_of_sales=models.PointOfSales.objects.first(),
            document_type__code=80,
        )
        factories.VatFactory(vat_type__code=5, receipt=receipt)
        factories.TaxFactory(tax_type__code=3, receipt=receipt)
        return receipt

    def test_validate_empty(self):
        factories.ReceiptFactory()

        errs = models.Receipt.objects.none().validate()

        self.assertEqual(errs, [])
        self.assertEqual(models.ReceiptValidation.objects.count(), 0)

    def test_validation_good(self):
        """Test validating valid receipts."""
        r1 = self._good_receipt()
        r2 = self._good_receipt()
        r3 = self._good_receipt()

        errs = models.Receipt.objects.all().validate()

        self.assertEqual(len(errs), 0)
        self.assertEqual(
            r1.validation.result,
            models.ReceiptValidation.RESULT_APPROVED,
        )
        self.assertEqual(
            r2.validation.result,
            models.ReceiptValidation.RESULT_APPROVED,
        )
        self.assertEqual(
            r3.validation.result,
            models.ReceiptValidation.RESULT_APPROVED,
        )
        self.assertEqual(models.ReceiptValidation.objects.count(), 3)

    def test_validation_bad(self):
        """Test validating invalid receipts."""
        self._bad_receipt()
        self._bad_receipt()
        self._bad_receipt()

        errs = models.Receipt.objects.all().validate()

        self.assertEqual(len(errs), 1)
        self.assertEqual(
            errs[0],
            "Error 10015: Factura B (CbteDesde igual a CbteHasta), DocTipo: "
            "80, DocNro 203012345 no se encuentra registrado en los padrones "
            "de AFIP y no corresponde a una cuit pais.",
        )
        self.assertQuerysetEqual(models.ReceiptValidation.objects.all(), [])

    def test_validation_mixed(self):
        """
        Test validating a mixture of valid and invalid receipts.

        Receipts are validated by AFIP in-order, so all receipts previous to
        the bad one are validated, and nothing else is even parsed after the
        invalid one.
        """
        r1 = self._good_receipt()
        self._bad_receipt()
        self._good_receipt()

        errs = models.Receipt.objects.all().validate()

        self.assertEqual(len(errs), 1)
        self.assertEqual(
            errs[0],
            "Error 10015: Factura B (CbteDesde igual a CbteHasta), DocTipo: "
            "80, DocNro 203012345 no se encuentra registrado en los padrones "
            "de AFIP y no corresponde a una cuit pais.",
        )
        self.assertQuerysetEqual(
            models.ReceiptValidation.objects.all(),
            [r1.pk],
            lambda rv: rv.receipt_id,
        )

    def test_validation_validated(self):
        """Test validating invalid receipts."""
        receipt = self._good_receipt()
        models.ReceiptValidation.objects.create(
            result=models.ReceiptValidation.RESULT_APPROVED,
            cae="123",
            cae_expiration=now(),
            receipt=receipt,
            processed_date=now(),
        )

        errs = models.Receipt.objects.all().validate()

        self.assertEqual(models.ReceiptValidation.objects.count(), 1)
        self.assertEqual(errs, [])

    def test_validation_good_service(self):
        """Test validating a receipt for a service (rather than product)."""
        receipt = self._good_receipt()
        receipt.concept_id = 2
        receipt.service_start = datetime.now() - timedelta(days=10)
        receipt.service_end = datetime.now()
        receipt.expiration_date = datetime.now() + timedelta(days=10)
        receipt.save()

        errs = models.Receipt.objects.all().validate()

        self.assertEqual(len(errs), 0)
        self.assertEqual(
            receipt.validation.result,
            models.ReceiptValidation.RESULT_APPROVED,
        )
        self.assertEqual(models.ReceiptValidation.objects.count(), 1)

    def test_validation_good_without_tax(self):
        """Test validating valid receipts."""
        receipt = factories.ReceiptFactory(
            point_of_sales=models.PointOfSales.objects.first(),
            total_amount=121,
        )
        factories.VatFactory(vat_type__code=5, receipt=receipt)

        errs = models.Receipt.objects.all().validate()

        self.assertEqual(len(errs), 0)
        self.assertEqual(
            receipt.validation.result,
            models.ReceiptValidation.RESULT_APPROVED,
        )
        self.assertEqual(models.ReceiptValidation.objects.count(), 1)

    def test_validation_good_without_vat(self):
        """Test validating valid receipts."""
        receipt = factories.ReceiptFactory(
            point_of_sales=models.PointOfSales.objects.first(),
            receipt_type__code=11,
            total_amount=109,
        )
        factories.TaxFactory(tax_type__code=3, receipt=receipt)

        errs = models.Receipt.objects.all().validate()

        self.assertEqual(len(errs), 0)
        self.assertEqual(
            receipt.validation.result,
            models.ReceiptValidation.RESULT_APPROVED,
        )
        self.assertEqual(models.ReceiptValidation.objects.count(), 1)

    @skip("Currently not working -- needs to get looked at.")
    def test_validation_with_observations(self):
        receipt = factories.ReceiptFactory(
            document_number=20291144404,
            document_type__code=80,
            point_of_sales=models.PointOfSales.objects.first(),
            receipt_type__code=1,
        )
        factories.VatFactory(vat_type__code=5, receipt=receipt)
        factories.TaxFactory(tax_type__code=3, receipt=receipt)

        errs = models.Receipt.objects.all().validate()

        self.assertEqual(len(errs), 0)
        self.assertEqual(
            receipt.validation.result,
            models.ReceiptValidation.RESULT_APPROVED,
        )
        self.assertEqual(models.ReceiptValidation.objects.count(), 1)
        self.assertEqual(models.Observation.objects.count(), 1)
        self.assertEqual(receipt.validation.observations.count(), 1)

    def test_credit_note(self):
        """Test validating valid a credit note."""
        # Create an invoice (code=6) and validate it...
        invoice = self._good_receipt()

        errs = models.Receipt.objects.filter(pk=invoice.pk).validate()
        assert len(errs) == 0
        assert models.ReceiptValidation.objects.count() == 1

        # Now create a credit note (code=8) and validate it...
        credit = self._good_receipt()
        credit.receipt_type = factories.ReceiptTypeFactory(code=8)
        credit.related_receipts.set([invoice])
        credit.save()

        errs = models.Receipt.objects.filter(pk=credit.pk).validate()
        assert len(errs) == 0
        assert models.ReceiptValidation.objects.count() == 2



import re
from datetime import date

import pytest
from django.test import TestCase

from django_afip import factories
from django_afip import models
from django_afip.pdf import ReceiptQrCode


class ReceiptPDFGenerationTestCase(TestCase):
    def test_pdf_generation(self):
        """Test PDF file generation.

        For the moment, this test case mostly verifies that pdf generation
        *works*, but does not actually validate the pdf file itself.

        Running this locally *will* yield the file itself, which is useful for
        manual inspection.
        """
        pdf = factories.ReceiptPDFFactory(receipt__receipt_number=3)
        factories.ReceiptValidationFactory(receipt=pdf.receipt)
        pdf.save_pdf()
        regex = r"afip/receipts/[a-f0-9]{2}/[a-f0-9]{2}/[a-f0-9]{32}.pdf"
        self.assertTrue(re.match(regex, pdf.pdf_file.name))
        self.assertTrue(pdf.pdf_file.name.endswith(".pdf"))

    def test_unauthorized_receipt_generation(self):
        """
        Test PDF file generation for unauthorized receipts.

        Confirm that attempting to generate a PDF for an unauthorized receipt
        raises.
        """
        taxpayer = factories.TaxPayerFactory()
        factories.TaxPayerProfileFactory(taxpayer=taxpayer)
        receipt = factories.ReceiptFactory(
            receipt_number=None,
            point_of_sales__owner=taxpayer,
        )
        pdf = models.ReceiptPDF.objects.create_for_receipt(
            receipt=receipt,
            client_name="John Doe",
            client_address="12 Green Road\nGreenville\nUK",
        )
        with self.assertRaisesMessage(
            Exception, "Cannot generate pdf for non-authorized receipt"
        ):
            pdf.save_pdf()


class GenerateReceiptPDFSignalTestCase(TestCase):
    def test_not_validated_receipt(self):
        printable = factories.ReceiptPDFFactory()

        self.assertFalse(printable.pdf_file)

    def test_validated_receipt(self):
        validation = factories.ReceiptValidationFactory()
        printable = factories.ReceiptPDFFactory(receipt=validation.receipt)

        self.assertTrue(printable.pdf_file)
        self.assertTrue(printable.pdf_file.name.endswith(".pdf"))


@pytest.mark.django_db
def test_qrcode_data():
    pdf = factories.ReceiptPDFFactory(
        receipt__receipt_number=3,
        receipt__issued_date=date(2021, 3, 2),
    )
    factories.ReceiptValidationFactory(receipt=pdf.receipt)

    qrcode = ReceiptQrCode(pdf.receipt)
    assert qrcode._data == {
        "codAut": "67190616790549",
        "ctz": 1.0,
        "cuit": "20329642330",
        "fecha": "2021-03-02",
        "importe": 130.0,
        "moneda": "PES",
        "nroCmp": 3,
        "nroDocRec": 203012345,
        "ptoVta": 1,
        "tipoCmp": 6,
        "tipoCodAut": "E",
        "tipoDocRec": 96,
        "ver": 1,
    }
